import unittest
import sys
import functools
import shutil
from typing import Union, Tuple

sys.path.insert(0, '../reticulator')
from reticulator import *

def get_packs() -> Tuple[BehaviorPack, ResourcePack]:
    project = Project('./content/bp/', './content/rp/')
    project.set_output_directory('out')
    return project.get_packs()

def save_and_return_packs(rp: ResourcePack = None, bp: BehaviorPack = None, force: bool = False):
    # Prepare folder location
    prepare_output_directory()

    # Save the old packs
    if rp is not None:
        rp.output_directory = './out/rp/'
        rp.save(force=force)
    
    if bp is not None:
        bp.output_directory = './out/bp/'
        bp.save(force=force)

    # Return the saved packs packs
    project = Project('./out/bp/', './out/rp/')
    return project.behavior_pack, project.resource_pack

def prepare_output_directory():
    try:
        shutil.rmtree('./out')
    except OSError as e:
        pass

    os.mkdir('./out')

## --------------- ##
## General Methods ##
## --------------- ##
class TestContextManager(unittest.TestCase):
    def test_context_manager(self):
        bp, rp = get_packs()

        with rp.get_entity('minecraft:dolphin') as dolphin:
            dolphin.set_jsonpath('test', 'test')

            # Dirty before save
            self.assertTrue(dolphin.dirty)

        # No longer dirty after implicit save (from context manager)
        self.assertFalse(dolphin.dirty)

class TestDirty(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.entity = self.bp.get_entity('minecraft:dolphin')
        self.function = self.bp.get_function('functions/kill_all_safe.mcfunction')
        self.command = self.function.commands[0]
        self.component = self.entity.get_component('minecraft:type_family')
        self.item_texture_file = self.rp.item_texture_file
        self.texture_definition = self.item_texture_file.get_texture_definition('axe')

    def assert_dirties(attribute):
        """
        Wrapper, which ensures that the resource is dirty, after the function is called.
        Attribute is passed in as a string, since `self` context is only available at runtime.
        """
        def _implementation(func):
            @functools.wraps(func)
            def wrapper(self, *args):
                resource = getattr(self, attribute)
                self.assertEqual(resource.dirty, False)
                func(self, *args)
                self.assertEqual(resource.dirty, True)
            return wrapper
        return _implementation


    @assert_dirties('function')
    def test_list_append(self):
        self.function.commands.append('a new command!')

    @assert_dirties('function')
    def test_list_delete(self):
        del self.function.commands[0]

    @assert_dirties('function')
    def test_list_edit(self):
        self.function.commands[0].data = 'new command'

    @assert_dirties('entity')
    def test_property(self):
        self.entity.identifier = 'bob'

    @assert_dirties('entity')
    def test_jsonpath(self):
        self.entity.set_jsonpath('new_key', {})

    @assert_dirties('component')
    @assert_dirties('entity')
    def test_subresource(self):
        self.component.set_jsonpath('new_key', "")

    @assert_dirties('component')
    @assert_dirties('entity')
    def test_subresource_id(self):
        self.component.id = 'new_component_name'

    @assert_dirties('item_texture_file')
    @assert_dirties('texture_definition')
    def test_texture_definition(self):
        self.texture_definition.shortname = 'new_shortname'

    @assert_dirties('item_texture_file')
    @assert_dirties('texture_definition')
    def test_texture_definition_textures(self):
        self.texture_definition.textures.append('new_texture')

    @assert_dirties('item_texture_file')
    def test_add(self):
        self.item_texture_file.add_texture_definition('new_definition', [])

class TestDeletion(unittest.TestCase):
    def setUp(self) -> None:
        self.project = Project('./content/bp/', './content/rp/')
        self.bp = self.project.behavior_pack
        self.rp = self.project.resource_pack

    def test_file_deletion(self):
        """
        Tests that we can delete a file.
        """

        # Delete entity
        entity = self.bp.get_entity('minecraft:dolphin')
        entity.delete()

        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)

        # Check that the file is gone
        self.assertIsNone(saved_bp.get_entity('minecraft:dolphin'))
            

    def test_subresource_deletion(self):
        # Delete entity
        entity = self.bp.get_entity('minecraft:dolphin')
        component = entity.get_component('minecraft:type_family')
        component.delete()

        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)
        entity = saved_bp.get_entity('minecraft:dolphin')

        self.assertIsNone(entity.get_component('minecraft:type_family'))
    
    def test_list_subresource_deletion(self):
        model = self.rp.get_model('geometry.dolphin')

        self.assertEqual(model.identifier, 'geometry.dolphin')
        self.assertEqual(len(model.bones), 9)
        bone = model.get_bone('bristle2')
        bone.delete()
        self.assertEqual(len(model.bones), 9)

        # This fails, because the list can no longer be accessed
        # saved_bp, saved_rp = save_and_return_packs(rp=self.rp)

        # bone = model.get_bone('bristle2')

class TestJsonPathAccess(unittest.TestCase):
    """
    Test various jsonpath access methods.
    """

    def setUp(self) -> None:
        self.project = Project('./content/bp/', './content/rp/')
        self.bp = self.project.behavior_pack
        self.rp = self.project.resource_pack
    
    def test_jsonpath_exists(self):
        """
        Test jsonpath_exists method.
        """

        entity = self.bp.get_entity('minecraft:dolphin')

        # Test exists
        self.assertTrue(entity.jsonpath_exists('minecraft:entity/description/identifier'))

        # Test does not exist
        self.assertFalse(entity.jsonpath_exists('dne'))

    def test_delete_jsonpath(self):
        """ 
        Tests deleting paths from json data.
        """
        entity = self.bp.get_entity('minecraft:dolphin')
        
        # Test deleting a path which does not exist
        entity.delete_jsonpath('dne')

        # Delete string
        entity.delete_jsonpath('minecraft:entity/description/identifier')
        self.assertFalse(entity.jsonpath_exists('minecraft:entity/description/identifier'))

        # Delete list
        path = 'minecraft:entity/events/minecraft:entity_spawned/randomize/0'

        # A complex test. Probably should be made clearer. 
        # The idea is that deleting a list from a jsonpath should set it to 
        # None, which the dpath lib apparently does by default
        self.assertEqual(len(entity.get_jsonpath('minecraft:entity/events/minecraft:entity_spawned/randomize')), 2)
        self.assertNotEqual(entity.get_jsonpath(path), None)
        entity.delete_jsonpath(path)
        self.assertEqual(entity.get_jsonpath(path), None)
        self.assertEqual(len(entity.get_jsonpath('minecraft:entity/events/minecraft:entity_spawned/randomize')), 2)

        # Delete complex structure
        entity.delete_jsonpath('minecraft:entity')
        self.assertFalse(entity.jsonpath_exists('minecraft:entity'))

    def test_set_jsonpath(self):
        """
        Test setting a value in a jsonpath.
        """
        entity = self.bp.get_entity('minecraft:dolphin')

        # Test normal set_jsonpath
        entity.set_jsonpath('new_key', 'new_value')
        self.assertEqual(entity.get_jsonpath('new_key'), 'new_value')

        # Test works if path does not exist
        entity.set_jsonpath('does_not_exist', 'new_value')

        # Test overwrite=False
        entity.set_jsonpath('minecraft:entity/description/identifier', 'minecraft:dog', overwrite=False)
        self.assertNotEqual(entity.get_jsonpath('minecraft:entity/description/identifier'), 'minecraft:dog')

        # Test overwrite=True
        entity.set_jsonpath('minecraft:entity/description/identifier', 'minecraft:dog', overwrite=True)
        self.assertEqual(entity.get_jsonpath('minecraft:entity/description/identifier'), 'minecraft:dog')

    def test_pop_jsonpath(self):
        """
        Tests popping a jsonpath
        """

        entity = self.bp.get_entity('minecraft:dolphin')

        # Test normal pop
        self.assertEqual(entity.pop_jsonpath('minecraft:entity/description/identifier'), 'minecraft:dolphin')

        # Test default
        self.assertEqual(entity.pop_jsonpath('dne', default='default_value'), 'default_value')

    def test_get_jsonpath(self):
        """
        Tests the result of a valid jsonpath.
        """

        # Set up
        dolphin = self.bp.get_entity('minecraft:dolphin')
        egg = self.bp.get_entity('minecraft:egg')

        # Test Property access
        self.assertEqual(dolphin.identifier, 'minecraft:dolphin')

        # Test jsonpath access
        self.assertEqual(
            len(dolphin.get_jsonpath('minecraft:entity/component_groups')), 7
        )

        # Test jsonpath which does not exist
        self.assertEqual(egg.component_groups, [])

        # Test jsonpath default
        self.assertEqual(dolphin.get_jsonpath('dne', default='default'), 'default')

        # Test jsonpath which does not exist
        with self.assertRaises(AssetNotFoundError):
            dolphin.get_jsonpath('dne')

class TestFormatVersion(unittest.TestCase):
    """
    Tests format version getters, errors and comparison
    """

    def setUp(self) -> None:
        self.project = Project('./content/bp/', './content/rp/')
        self.bp = self.project.behavior_pack
        self.rp = self.project.resource_pack
        self.entity = self.bp.get_entity('minecraft:dolphin')
        self.recipe = self.bp.get_recipe('minecraft:acacia_boat')
        self.loot_table = self.bp.get_loot_table('loot_tables/entities/dolphin.json')

    def test_format_version(self):
        # Test getters
        self.assertEqual(self.entity.format_version, '1.16.0')
        self.assertEqual(self.recipe.format_version, '1.12.0')
        with self.assertRaises(AttributeError):
            self.loot_table.format_version

        # Test Initaliser
        self.assertEqual(FormatVersion('1'), '1.0.0.0')
        self.assertEqual(FormatVersion('12.4'), FormatVersion(FormatVersion('12.4.0.3.4')))

        # Test comparison
        self.assertTrue(self.entity.format_version > self.recipe.format_version)

        # Test setter
        self.entity.format_version = '1.17.0'
        self.assertEqual(self.entity.format_version, FormatVersion('1.17'))

## --------------------- ##
## Behavior Pack Classes ##
## --------------------- ##

class TestAnimationControllerBP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.animation_controller_file = self.bp.get_animation_controller_file('animation_controllers/example.ac.json')

    def test_animation_controller_files(self):
        self.assertEqual(len(self.bp.animation_controller_files), 2)

        self.bp.get_animation_controller_file('animation_controllers/example.ac.json')
        self.assertIsNone(self.bp.get_animation_controller_file('animation_controllers/dne.json'))
            

    def test_animation_controllers(self): 
        # From pack
        self.assertEqual(len(self.bp.animation_controllers), 3)
        # From animation controller file
        self.assertEqual(len(self.animation_controller_file.animation_controllers), 2)

        self.bp.get_animation_controller('controller.animation.test')
        self.assertIsNone(self.bp.get_animation_controller('controller.animation.dne'))
            

    def test_add_animation_controller_file(self): pass

    def test_add_animation_controller(self): pass

##TODO: Need to add class
class TestAnimationBP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.animation_file = self.bp.get_animation_file('animations/test.a.json')
        
    def test_animation_files(self):
        self.assertEqual(len(self.bp.animation_files), 1)
        self.bp.get_animation_file('animations/test.a.json')
        self.assertIsNone(self.bp.get_animation_file('animations/dne.json'))
            

    def test_animations(self): 
        # From pack
        self.assertEqual(len(self.bp.animations), 1)
        # From animation file
        self.assertEqual(len(self.animation_file.animations), 1)

        self.bp.get_animation('animation.test')
        self.assertIsNone(self.bp.get_animation('animation.dne'))
           

    def test_add_animation_file(self): pass

    def test_add_animation(self): pass

##TODO: Implement class

class TestBlockFileBP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.block = self.bp.get_block('namespace:block')
        
    def test_blocks(self): 
        self.assertEqual(len(self.bp.blocks), 1)

        self.bp.get_block('namespace:block')
        self.assertIsNone(self.bp.get_block('namespace:dne'))
            
     
    def test_add_block(self): pass

    def test_block_properties(self): 
        self.assertEqual(self.block.format_version, "1.10.0")
        self.assertEqual(self.block.identifier, "namespace:block")

        self.assertEqual(len(self.block.components), 1)
        self.block.get_component('minecraft:destroy_time')
        
        self.assertIsNone(self.block.get_component('minecraft:dne'))
            

        self.block.add_component(id="minecraft:display_name", data="Block")

        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)

        block = saved_bp.get_block('namespace:block')
        self.assertEqual(len(block.components), 2)

"""
class TestBiomeFile(unittest.TestCase):
    def test_biomes(self): pass

    def test_get_biome_file(self): pass

    def test_add_biome_file(self): pass
"""

class TestEntityFileBP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.entity = self.bp.get_entity('minecraft:dolphin')

    def test_entities(self): 
        self.assertEqual(len(self.bp.entities), 2)
        self.bp.get_entity('minecraft:dolphin')

        self.assertIsNone(self.bp.get_entity('minecraft:dne'))

    def test_add_entity_bp(self): pass

    def test_entity_bp_properties(self):
        self.assertEqual(self.entity.identifier,'minecraft:dolphin')

    def test_component_groups(self):
        self.assertEqual(len(self.entity.component_groups), 7)

        self.group = self.entity.get_component_group('dolphin_adult')
        self.assertEqual(self.group.id, 'dolphin_adult')
        self.assertEqual(len(self.group.components), 4)
        
    def test_add_component_groups(self):
        # Original Number
        self.assertEqual(len(self.entity.component_groups), 7)

        # Test adding component group by name and data
        component_group_data = { "minecraft:damage" : { "value" : 1 } }
        component_group = self.entity.add_component_group(id='group:one', data=component_group_data)
        self.assertEqual(component_group.id, 'group:one')
        self.assertEqual(len(self.entity.component_groups), 8)

        # Test adding component group by class
        group = ComponentGroup(data={}, parent=self.entity, json_path='minecraft:entity/component_groups/group:two')

        added_group = self.entity.add_component_group(resource=group)
        self.assertEqual(added_group.id, 'group:two')
        self.assertEqual(len(self.entity.component_groups), 9)

        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)

        # Test after saving
        saved_entity = saved_bp.get_entity('minecraft:dolphin')

        new_group = saved_entity.get_component_group('group:one')
        self.assertEqual(new_group.id, 'group:one')
        self.assertEqual(len(new_group.components), 1)

    def test_add_component(self):
        # Tests before
        self.assertEqual(len(self.entity.components), 29)

        # Add to components
        component_data = { "value" : 1 }
        self.entity.add_component(id="minecraft:damage", data=component_data)

        saved_bp, saved_rp = save_and_return_packs(bp=self.bp)

        saved_entity = saved_bp.get_entity('minecraft:dolphin')
        self.assertEqual(len(saved_entity.components), 30)

    def test_events(self): pass

    def test_get_event(self): pass

    def test_add_event(self): pass

class TestFeatureFile(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_feature_files(self): pass

    def test_add_feature_file(self): pass

class TestFeatureRuleFile(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_feature_rule_files(self): pass

    def test_add_feature_rule_file(self): pass

class TestFunctions(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_functions(self): 
        self.assertEqual(len(self.bp.functions), 2)

        # Getting function by path can take multiple path formats
        self.assertTrue(self.bp.get_function('functions/kill_all_safe.mcfunction'))
        self.assertTrue(self.bp.get_function('functions/teleport/home.mcfunction'))

        self.assertIsNone(self.bp.get_function('functions/no_function.mcfunction'))
            

    def test_add_function(self): pass

    def test_commands(self): 
        self.function = self.bp.get_function('functions/kill_all_safe.mcfunction')
        self.assertEqual(len(self.function.commands), 4)

    def test_set_command(self):
        self.function = self.bp.get_function('functions/kill_all_safe.mcfunction')
        command = self.function.commands[0]

        self.assertEqual(command.data, '# Remove all entities except players')
        command.data = 'new'
        self.assertEqual(command.data, 'new')

    def test_comment_stripping(self):
        """
        The first function has 1 command, the second has 2
        """

        # With stripping off
        self.assertEqual(len(self.bp.functions[0].commands), 4)
        self.assertEqual(len(self.bp.functions[1].commands), 2)

        # With stripping on
        self.bp.functions[0].strip_comments() # Strips 2 comments from the first function
        self.assertEqual(len(self.bp.functions[0].commands), 2) 

class TestItemFileBP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_items(self): pass

    def test_add_item(self): pass

    def test_components(self): pass

    def test_add_component(self): pass

class TestLootTables(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_loot_tables(self):
        self.assertEqual(len(self.bp.loot_tables), 2)
        self.assertEqual(self.bp.loot_tables[0].file_name, 'dolphin.json')
    
    def test_get_loot_table(self):
        dolphin = self.bp.get_entity('minecraft:dolphin')
        group = dolphin.get_component_group('dolphin_adult')
        component = group.get_component('minecraft:loot')
        table_name = component.data['table']
        loot_table = self.bp.get_loot_table(table_name)
        self.assertEqual(len(loot_table.pools), 1)

    def test_add_loot_table(self): pass

    def test_pools(self):
        self.assertEqual(len(self.bp.loot_tables[0].pools), 1)

    def test_get_pool(self): pass

    def test_add_pool(self): pass
      
class TestRecipes(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_recipes(self):
        self.assertEqual(len(self.bp.recipes), 5)

    def test_get_recipe_file(self):
        """
        Test all possible recipe types.
        """
        self.assertEqual(self.bp.get_recipe("minecraft:acacia_boat").identifier, "minecraft:acacia_boat")
        self.assertEqual(self.bp.get_recipe("minecraft:andesite").identifier, "minecraft:andesite")
        self.assertEqual(self.bp.get_recipe("minecraft:brew_awkward_blaze_powder").identifier, "minecraft:brew_awkward_blaze_powder")
        self.assertEqual(self.bp.get_recipe("minecraft:brew_splash_potion_dragon_breath").identifier, "minecraft:brew_splash_potion_dragon_breath")
        self.assertEqual(self.bp.get_recipe("minecraft:furnace_stained_hardened_clay3").identifier, "minecraft:furnace_stained_hardened_clay3")
        self.assertIsNone(self.bp.get_recipe("dne"))


    def test_add_recipe_file(self): pass

##class TestScripts(unittest.TestCase):pass

class TestSpawnRuleFile(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_spawn_rules(self): pass

    def test_add_spawn_rule(self): pass

##class TestTradeTables(unittest.TestCase):pass

## --------------------- ##
## Resource Pack Classes ##
## --------------------- ##
class TestAnimationControllerRP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_animation_controllers_files(self): pass

    def test_animation_controllers(self): pass

    def test_animation_controller_properties(self): pass

    def test_add_animation_controller_file(self): pass

    def test_add_animation_controller(self): pass

    def test_states(self): pass

    def test_add_state(self): pass

class TestAnimationRP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.animation_file = self.rp.get_animation_file('animations/dolphin.animation.json')

    def test_animation_files(self):
        self.assertEqual(1, len(self.rp.animation_files))

        self.assertTrue(self.rp.get_animation_file('animations/dolphin.animation.json'))
    
    def test_animations(self): 
        self.assertEqual(1, len(self.animation_file.animations))

        self.assertTrue(self.animation_file.get_animation('animation.dolphin.move'))

    def test_animation_file_properties(self): pass       

    def test_add_animation_file(self): pass

    def test_add_animation(self): pass

    def test_bones(self): pass

    def test_get_bone(self): pass

    def test_add_bone(self): pass

class TestAttachable(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_attachables(self): pass

    def test_add_attachable(self): pass

class TestEntityFileRP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.entity = self.rp.get_entity('minecraft:dolphin')

    def test_entities(self): pass

    def test_add_entity_rp(self): pass

    def test_entity_rp_properties(self): pass

    def test_animations(self): pass

    def test_models(self): pass

    def test_textures(self): pass

    def test_materials(self): pass

    # Getting & Adding handled in each class

class TestFogs(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.fog_file = self.rp.get_fog('minecraft:fog_mushroom_island_shore')

    def test_fog_files(self):
        self.assertEqual(len(self.rp.fogs), 3)
        self.rp.get_fog('minecraft:fog_mushroom_island_shore')

    def test_add_fog_file(self): pass

    def test_fog_file_properties(self): 
        self.assertEqual(len(self.fog_file.distance_components), 2)
        self.assertEqual(len(self.fog_file.volumetric_density_components), 1)
        self.assertEqual(len(self.fog_file.volumetric_media_coefficients), 2)

        component = self.fog_file.get_distance_component('air')
        self.assertEqual(component.data['fog_end'], 60)

class TestItemFileRP(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        
    def test_items(self): pass

    def test_add_item_rp(self): pass

    def test_item_rp_properties(self): pass

class TestMaterials(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_material_files(self):
        self.assertEqual(len(self.rp.material_files), 1)
        self.rp.get_material_file('materials/test.material')
        self.assertIsNone(self.rp.get_material_file('materials/dne.material'))
            

    def test_materials(self):
        self.assertEqual(len(self.rp.materials), 5)

        self.rp.get_material('dolphin')
        self.assertIsNone(self.rp.get_material('dne'))
            

    def test_add_material_file(self): pass

    def test_add_material(self): pass

    def test_material_file_properties(self): pass

    def test_material_properties(self): pass

class TestMaterialTriple(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.dolphin = self.rp.get_entity('minecraft:dolphin')
        self.elder_guardian = self.rp.get_entity('minecraft:elder_guardian')

    def test_existing_resources(self):
        """
        Tests that we can generate and use a ShortnameResourceTriple
        """

        materials = self.dolphin.materials
        self.assertEqual(len(materials), 1)

        material = materials[0]
        self.assertEqual(material, self.dolphin.get_material('default'))
        self.assertEqual(material.shortname, 'default')
        self.assertEqual(material.resource.id, 'dolphin')
        self.assertEqual(material.identifier, 'dolphin')

    def test_missing_resources(self):
        """
        Test the behavior of ShortnameResourceTriple when the resource is missing.
        """

        materials = self.elder_guardian.materials
        self.assertEqual(len(materials), 2)

        # Get the last material, which is missing
        material = materials[1]

        self.assertEqual(material.shortname, 'ghost')
        self.assertEqual(material.identifier, 'guardian_ghost')
        self.assertIsNone(material.resource)
            
    
class TestModels(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_model(self): 
        self.assertEqual(len(self.rp.models), 1)

    def test_model_files(self): 
        self.assertTrue(len(self.rp.model_files), 1)

    def test_add_model_file(self):pass

    def test_add_model(self):pass

    def test_model_properties(self):
        model = self.rp.get_model('geometry.dolphin')
        self.assertEqual(len(model.bones), 9)

class TestParticle(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_particles(self):pass

    def test_add_partcile(self):pass

    def test_particle_properties(self):
        """
        Checks that the properties of the test particle is correct.
        """
        particle = self.rp.get_particle('minecraft:death_explosion_emitter')
        self.assertEqual(particle.identifier, 'minecraft:death_explosion_emitter')
        self.assertEqual(particle.format_version, "1.10.0")

    def test_particle_path(self):
        """
        Checks whether the FileResource properties are correct.
        """
        self.assertEqual(len(self.rp.particles), 2)
        particle = self.rp.get_particle('minecraft:death_explosion_emitter')
        self.assertEqual(particle.file_name, 'explosion_death.json')
        self.assertEqual(particle.filepath, 'particles\\explosions\\explosion_death.json')

    def test_particle_subresources(self):
        """
        Tests access into the 'components' property of the particle.
        """ 
        particle = self.rp.get_particle('minecraft:death_explosion_emitter')
        self.assertEqual(len(particle.components), 10)
        component = particle.get_component('minecraft:emitter_rate_instant')
        self.assertEqual(component.id, 'minecraft:emitter_rate_instant')
        self.assertEqual(component.get_jsonpath('num_particles'), 20)
        self.assertEqual(component.data['num_particles'], 20)

class TestRenderControllers(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_render_controller_files(self):
        self.assertEqual(len(self.rp.render_controller_files), 1)

        self.rp.get_render_controller_file('render_controllers/dolphin.render_controller.json')

    def test_render_controllers(self):
        self.assertEqual(len(self.rp.render_controllers), 2)

        self.rp.get_render_controller('controller.render.dolphin')

    def test_add_render_controller_file(self): pass

    def test_add_render_controller(self):
        rcf = self.rp.get_render_controller_file('render_controllers/dolphin.render_controller.json')

        # Original length
        self.assertEqual(len(self.rp.render_controllers), 2)
        rc = rcf.add_render_controller(id='controller.render.test', data={})

        # After adding the render controller
        self.assertEqual(len(self.rp.render_controllers), 3)
    
    def test_render_controller_file_properties(self):
        self.assertEqual(self.rp.get_render_controller('controller.render.dolphin').file.format_version, '1.8.0')

    def test_render_controller_properties(self):
        self.assertEqual(self.rp.get_render_controller('controller.render.dolphin').file.format_version, '1.8.0')

class TestSounds(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
    
    def test_sounds(self):
        self.assertEqual(len(self.rp.sounds), 2)
        self.assertEqual(self.rp.sounds[0], 'sounds/bottle/fill_dragonbreath1.fsb')
    
    def test_sounds_file(self):
        with self.assertRaises(AssetNotFoundError):
            self.rp.sounds_file
        
    def test_sound_definitions_file(self):
        self.assertEqual(self.rp.sound_definitions_file.format_version, '1.10.0')

class TestTextures(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_textures(self):
        self.assertEqual(len(self.rp.textures), 5)
        self.assertEqual(self.rp.textures[0], 'textures/blocks/ancient_debris_top.png')

    def test_get_textures(self):
        self.assertEqual(len(self.rp.get_textures('entity')), 2)
        self.assertEqual(len(self.rp.get_textures('dne')), 0)
        self.assertEqual(len(self.rp.get_textures('')), 5)

    def test_get_textures_trim(self):
        self.assertEqual(self.rp.get_textures('entity', trim_extension=False)[0], 'textures/entity/alex.png')
        self.assertEqual(self.rp.get_textures('entity', trim_extension=True)[0], 'textures/entity/alex')

class TestStandaloneTextureFiles(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
    
    def test_terrain_texture_file(self):
        terrain_texture = self.rp.terrain_texture_file
        texture_definition = terrain_texture.texture_definitions[0]

        self.assertEqual(len(terrain_texture.texture_definitions), 5)
        self.assertEqual(texture_definition.textures[0], 'textures/blocks/planks_acacia')

        texture_definition = terrain_texture.get_texture_definition('anvil_base')

        self.assertEqual(texture_definition.textures[0], 'textures/blocks/anvil_base')
        self.assertEqual(len(texture_definition.textures), 4)

    def test_item_texture_file(self):
        item_texture_file = self.rp.item_texture_file
        texture_definition = item_texture_file.texture_definitions[0]

        self.assertEqual(len(item_texture_file.texture_definitions), 5)
        self.assertEqual(texture_definition.textures[0], 'textures/items/apple')
    
        texture_definition = item_texture_file.get_texture_definition('axe')

        self.assertEqual(texture_definition.textures[0], 'textures/items/wood_axe')
        self.assertEqual(len(texture_definition.textures), 6)

    def test_flipbook_texture_file(self):pass



class TestBiomesClientFile(unittest.TestCase):pass

class TestBlocksFile(unittest.TestCase):pass

## --------------- ##
## General Classes ##
## --------------- ##
class TestAnimationTriple(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()
        self.dolphin = self.rp.get_entity('minecraft:dolphin')
        self.elder_guardian = self.rp.get_entity('minecraft:elder_guardian')

    def test_existing_resources(self):
        """
        Tests that we can generate and use a ShortnameResourceTriple
        """

        animations = self.dolphin.animations
        self.assertEqual(len(animations), 1)

        animation = animations[0]
        self.assertEqual(animation.shortname, 'move')
        self.assertEqual(animation.resource.id, 'animation.dolphin.move')
        self.assertEqual(animation.identifier, 'animation.dolphin.move')

    def test_missing_resources(self):
        """
        Test the behavior of ShortnameResourceTriple when the resource is missing.
        """

        animations = self.elder_guardian.animations
        self.assertEqual(len(animations), 5)

        # Get the last animation, which is missing
        animation = animations[-1]

        self.assertEqual(animation.shortname, 'missing')
        self.assertEqual(animation.identifier, 'animation.guardian.missing')
        self.assertIsNone(animation.resource)

    def test_saving(self):
        """
        Tests that we can save AnimationTriple
        """

        # Edit the resource
        animation : AnimationTriple = self.dolphin.animations[0]
        animation.shortname = 'new_shortname'
        animation.identifier = 'new_identifier'

        # Save the resource
        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)

        animation = saved_rp.get_entity('minecraft:dolphin').animations[0]

        self.assertEqual(animation.shortname, 'new_shortname')
        self.assertEqual(animation.identifier, 'new_identifier')

        # Raise error for the renamed resource
        with self.assertRaises(AssetNotFoundError):
            self.dolphin.get_animation('move')

class TestLanguageFiles(unittest.TestCase):
    def setUp(self) -> None:
        self.bp, self.rp = get_packs()

    def test_language_file(self):
        self.assertEqual(len(self.rp.language_files), 1)

        language_file = self.rp.get_language_file('texts/es_ES.lang')

        with self.assertRaises(AssetNotFoundError):
            self.rp.get_language_file('texts/dne.lang')

        self.assertEqual(len(language_file.translations), 3)
        translation = language_file.get_translation('accessibility.text.period')

        self.assertEqual(translation.key, 'accessibility.text.period')
        self.assertEqual(translation.value, 'Punto')
        self.assertEqual(translation.comment, '')

    def test_adding_translation(self):
        language_file = self.rp.get_language_file('texts/es_ES.lang')
        language_file.add_translation(Translation('new_key', 'new_value'))
        language_file.add_translation(Translation('new_key2', 'new_value2', comment="Test"))

        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)

        language_file = saved_rp.get_language_file('texts/es_ES.lang')
        self.assertEqual(len(language_file.translations), 5)
        translation = language_file.get_translation('new_key2')
        self.assertEqual(translation.comment, 'Test')

    def test_overwrite_translation(self):
        language_file = self.rp.get_language_file('texts/es_ES.lang')
        language_file.add_translation(Translation('accessibility.text.period', 'Test 1'),overwrite=True)
        language_file.add_translation(Translation('accessibility.text.comma', 'Test 2'),overwrite=False)

        saved_bp, saved_rp = save_and_return_packs(rp=self.rp)

        language_file = saved_rp.get_language_file('texts/es_ES.lang')
        self.assertEqual(len(language_file.translations), 3)
        translation_1 = language_file.get_translation('accessibility.text.period')
        translation_2 = language_file.get_translation('accessibility.text.comma')
        self.assertEqual(translation_1.value,'Test 1')
        self.assertEqual(translation_2.value,'Coma')





if __name__ == '__main__':
    unittest.main()


from __future__ import annotations

import re
import os
import json
import glob
import functools 

from pathlib import Path
from functools import cached_property
from typing import Union, Tuple, TypeVar, Any

import dpath

NO_ARGUMENT = object()
TEXTURE_EXTENSIONS = ["png", "jpg", "jpeg", "tga"]
SOUND_EXTENSIONS = ["wav", "fsb", "ogg"]

T = TypeVar('T')

class TypeInfo:
    """
    Class for encapsulating the arguments that are passed to sub-resource based
    decorators.
    """
    
    def __init__(self, *, jsonpath="", filepath="", attribute="", getter_attribute="", extension=".json"):
        self.jsonpath = jsonpath
        self.filepath = filepath
        self.attribute = attribute
        self.getter_attribute = getter_attribute
        self.extension = extension


def convert_to_notify_structure(data: Union[dict, list], parent: Resource) -> Union[NotifyDict, NotifyList]:
    """
    Converts a dict or list to a notify structure.
    """
    if isinstance(data, dict):
        return NotifyDict(data, owner=parent)

    if isinstance(data, list):
        return NotifyList(data, owner=parent)

    return data


def format_version(jsonpath: str = "format_version"):
    """
    Class Decorator which inserts a 'format_version' property, with proper
    semantics.
    """

    def inner_format_version(cls):
        @property
        def format_version(self) -> FormatVersion:
            return FormatVersion(self.get_jsonpath(jsonpath))
        
        @format_version.setter
        def format_version(self, format_version):
            return self.set_jsonpath(jsonpath, str(FormatVersion(format_version)))

        cls.format_version = format_version
        return cls
    
    return inner_format_version


def identifier(jsonpath: str):
    """
    Class Decorator, which injects handling for accessing 'identifier'.
    :param jsonpath: The jsonpath where the identifier can be located.
    """

    def inner(cls):
        @property
        def identifier(self) -> str:
            return self.get_jsonpath(jsonpath)
        
        @identifier.setter
        def identifier(self, identifier):
            return self.set_jsonpath(jsonpath, identifier)

        cls.identifier = identifier
        
        return cls

    return inner


def ClassProperty(attribute: str, jsonpath: str = NO_ARGUMENT):
    """
    Class Decorator which injects a single 'property' with proper semantics.

    :param attribute: The attribute where this property can be accessed from
    :param jsonpath: The jsonpath where this attribute can be found in the json data
    """

    # Allow single-argument class-properties
    if jsonpath == NO_ARGUMENT:
        jsonpath = attribute

    def inner(cls):
        @property
        def template_property(self) -> str:
            return self.get_jsonpath(jsonpath)
        
        @template_property.setter
        def template_property(self, new_prop):
            return self.set_jsonpath(jsonpath, new_prop)

        setattr(cls, attribute, template_property)

        return cls
    return inner


def JsonChildResource(parent_cls: Resource, child_cls: T):
    """
    Special case, for looping over all children FILES to access
    RESOURCES. 
    
    Classic example is doing `rp.animations` instead of looping over all
    animation files and then painfully getting animations from there.

    NOT CACHED
    """
    
    parent_attribute = parent_cls.type_info.attribute
    child_attribute = child_cls.type_info.attribute

    def decorator(func) -> cached_property[List[T]]:
        @property
        @functools.wraps(func)
        def wrapper(self) -> List[T]:
            sub_resources = []
            for file_resource in getattr(self, parent_attribute):
                for sub_resource in getattr(file_resource, child_attribute):
                    sub_resources.append(sub_resource)
            return sub_resources
        return wrapper
    return decorator


def ResourceDefinition(cls : T):
    """Inserts implementation for JsonFileResource"""

    attribute = cls.__name__
    filepath = cls.type_info.filepath
    extension = cls.type_info.extension

    def decorator(func) -> cached_property[list[T]]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> list[T]:
            setattr(self, attribute, [])
            base_directory = os.path.join(self.input_path, filepath)

            for local_path in glob.glob(base_directory + "/**/*" + extension, recursive=True):
                local_path = os.path.relpath(local_path, self.input_path)

                getattr(self, attribute).append(cls(filepath = local_path, pack = self))
            return getattr(self, attribute)
        return wrapper
    return decorator


def SingleResourceDefinition(cls : T):
    """Inserts implementation for JsonFileResource (single)"""

    attribute = cls.__name__
    filepath = cls.type_info.filepath
    extension = cls.type_info.extension

    def decorator(func) -> cached_property[T]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> T:
            new_object = cls(filepath = filepath + extension, pack = self)
            setattr(self, attribute, new_object)
            return new_object
        return wrapper
    return decorator


def SubResourceDefinition(cls: T):
    jsonpath = cls.type_info.jsonpath
    attribute = cls.type_info.attribute
    def decorator(func) -> cached_property[list[T]]:
        @cached_property
        @functools.wraps(func)
        def wrapper(self) -> list[T]:
            setattr(self, attribute, [])
            for path, data in self.get_data_at(jsonpath):
                getattr(self, attribute).append(cls(parent = self, json_path = path, data = data))
            return getattr(self, attribute)
        return wrapper
    return decorator


def Getter(cls : T):
    """
    Decorator which allows you to get a resource. For example getting a component
    from an entity.
    """
    attribute = cls.type_info.attribute
    getter_attribute = cls.type_info.getter_attribute
    def decorator(func) -> T:
        @functools.wraps(func)
        def wrapper(self, compare):
            for child in getattr(self, attribute):
                if smart_compare(getattr(child, getter_attribute), compare):
                    return child
            return None # No longer raises an error. Allow a getter to return none.
        return wrapper
    return decorator


def ChildGetter(parent_cls : Any, child_cls : T):
    """
    Special getter for getting sub-resources that are further down the chain.
    For example getting 'animations' directly from the RP without passing
    through the AnimationFile class.
    """
    parent_attribute = parent_cls.type_info.attribute
    child_attribute = child_cls.type_info.attribute
    getter_attribute = child_cls.type_info.getter_attribute
    def decorator(func) -> T:
        @functools.wraps(func)
        def wrapper(self, compare):
            for child in getattr(self, parent_attribute):
                for grandchild in getattr(child, child_attribute):
                    if smart_compare(getattr(grandchild, getter_attribute), compare):
                        return grandchild
                return None
        return wrapper
    return decorator
    
def SubResourceAdder(cls : Resource):
    """
    This decorator allows you to inject SubResources into your Resources.
    """
    jsonpath = cls.type_info.jsonpath
    attribute = cls.type_info.attribute

    def decorator_sub_resource(func):
        @functools.wraps(func)
        def wrapper_sub_resource(self, *args, **kwargs):
            # This handles the case where a class is passed in directly.
            # like `add_box(Box(..))``

            if len(args) > 0:
                raise ReticulatorException("This function can only be called with keyword arguments: 'resource' (alone), or 'id' and 'data'")

            if kwargs.get('resource') == None and (kwargs.get('id') == None or kwargs.get('data') == None):
                raise ReticulatorException("This function can only be called with 'resource' OR 'id and 'data'.")

            # Handle Object case
            if new_object := kwargs.get('resource'):
                self.set_jsonpath(new_object.json_path, new_object.data)
                new_object.parent = self

            # This handles the 'normal' flow, where arguments are passed in
            # via 'parts' and are constructed automatically.
            else:
                id = kwargs.get('id')
                data = kwargs.get('data')

                if id == None:
                    raise ReticulatorException("Id may not be None.")

                if data == None:
                    raise ReticulatorException("Data may not be None")

                new_jsonpath = jsonpath + "/" + id
                self.set_jsonpath(new_jsonpath, data)
                new_object = cls(data=data, parent=self, json_path=new_jsonpath)

            if new_object:
                getattr(self, attribute).append(new_object)
                return new_object
            else:
                raise ReticulatorException()


        return wrapper_sub_resource
    return decorator_sub_resource

# def ResourceAdder(cls : Resource):
#     filepath = cls.type_info.filepath
#     attribute = cls.type_info.attribute

#     def decorator_sub_resource(func):
#         @functools.wraps(func)
#         def wrapper_sub_resource(self, *args):
#             # This handles the case where a class is passed in directly.
#             # like `add_box(Box(..))``
#             if len(args) == 1 and isinstance(args[0], cls):
#                 args[0].pack = self

#             # This handles the 'normal' flow, where arguments are passed in
#             # via 'parts' and are constructed automatically.
#             else:
#                 name = args[0]
#                 data = args[1]

#                 new_jsonpath = jsonpath + "/" + name
#                 self.set_jsonpath(new_jsonpath, data)
#                 new_object = cls(data=data, parent=self, json_path=new_jsonpath)
            
#             # Last step is adding the new object to the attribute
#             if new_object:
#                 getattr(self, attribute).append(new_object)
#                 return new_object
#             else:
#                 raise ReticulatorException()


#         return wrapper_sub_resource
#     return decorator_sub_resource

# Methods
def create_nested_directory(path: str):
    """
    Creates a nested directory structure if it doesn't exist.
    """
    Path(path).parent.mkdir(parents=True, exist_ok=True)

def save_json(filepath, data):
    """
    Saves json to a filepath, creating nested directory if required.
    """
    create_nested_directory(filepath)
    with open(filepath, "w+") as file_head:
        return json.dump(data, file_head, indent=2, ensure_ascii=False)

def smart_compare(a, b) -> bool:
    """
    Compares to objects using ==, but if they can both be interpreted as
    path-like objects, it uses path comparison.
    """

    try:
        a = Path(a)
        b = Path(b)

        return a == b
    except Exception:
        return a == b


# Exceptions
class ReticulatorException(Exception):
    """
    Base class for Reticulator exceptions.
    """

class FloatingAssetError(ReticulatorException):
    """
    Called when a "floating" asset attempts to access its parent pack, for
    example when saving
    """

class AssetNotFoundError(ReticulatorException):
    """
    Called when attempting to access an asset that does not exist, for
    example getting an entity by name
    """

class InvalidJsonError(ReticulatorException):
    """
    Called when a JSON file is invalid.
    """

class AmbiguousAssetError(ReticulatorException):
    """
    Called when a path is not unique.
    """

class FormatVersionError(ReticulatorException):
    """
    Called when a format version does not exist for a file
    """


# Base Classes

class NotifyDict(dict):
    """
    A notify dictionary is a dictionary that can notify its parent when its been
    edited.
    """
    def __init__(self, *args, owner: Resource = None, **kwargs):
        self._owner = owner

        if len(args) > 0:
            for key, value in args[0].items():
                args[0][key] = convert_to_notify_structure(value, self._owner)
        super().__init__(*args, **kwargs)

    def get_item(self, attr):
        try:
            return self.__getitem__(attr)
        except Exception:
            return None
    
    def __delitem__(self, v) -> None:
        if(self._owner != None):
            self._owner.dirty = True
        return super().__delitem__(v)

    def __setitem__(self, attr, value):
        value = convert_to_notify_structure(value, self._owner)

        if self.get_item(attr) != value:
            if(self._owner != None):
                self._owner.dirty = True

        super().__setitem__(attr, value)

class NotifyList(list):
    """
    A notify list is a list which can notify its owner when it has
    changed.
    """
    def __init__(self, *args, owner: Resource = None, **kwargs):
        self._owner = owner

        if len(args) > 0:
            for i in range(len(args[0])):
                args[0][i] = convert_to_notify_structure(args[0][i], self._owner)

        super().__init__(*args, **kwargs)

    def get_item(self, attr):
        try:
            return self.__getitem__(attr)
        except Exception:
            return None
    
    def __delitem__(self, v) -> None:
        if(self._owner != None):
            self._owner.dirty = True
        return super().__delitem__(v)
    
    def append(self, v):
        if(self._owner != None):
            self._owner.dirty = True
        super().append(v)

    def extend(self, v):
        if(self._owner != None):
            self._owner.dirty = True
        super().extend(v)

    def __setitem__(self, attr, value):
        value = convert_to_notify_structure(value, self._owner)

        if self.__getitem__(attr) != value:
            if(self._owner != None):
                self._owner.dirty = True
        
        super().__setitem__(attr, value)

class Pack():
    """
    Pack is a parent class that contains the shared functionality between the
    resource pack and the behavior pack.
    """
    def __init__(self, input_directory: str, project : Project = None):
        self.resources = []
        self._language_files = []
        self._project = project

        # The input path is the path to the folder containing the pack.
        self.input_path: str = input_directory

        # The output path is the path to the output directory, where this
        # pack will be saved. This is the same as the input path, unless
        # explicitely saved.
        self.output_directory: str = input_directory

    @cached_property
    def project(self) -> Project:
        """
        Get the project that this pack is part of. May be None.
        """
        return self._project

    def save(self, force=False):
        """
        Saves every child resource.
        """
        for resource in self.resources:
            resource.save(force=force)

    def register_resource(self, resource: Resource) -> None:
        """
        Register a child resource to this pack. These resources will be saved
        during save.
        """

        # Attempt to set the pack for this resource.
        try:
            resource.pack = self
        except Exception:
            pass
        self.resources.append(resource)

    def get_language_file(self, filepath:str) -> LanguageFile:
        """
        Gets a specific language file, based on the name of the language file.
        For example, 'texts/en_GB.lang'
        """
        for language_file in self.language_files:
            if smart_compare(language_file.filepath, filepath):
                return language_file
        raise AssetNotFoundError(filepath)

    @cached_property
    def language_files(self) -> list[LanguageFile]:
        """
        Returns a list of LanguageFiles, as read from 'texts/*'
        """
        base_directory = os.path.join(self.input_path, "texts")
        for local_path in glob.glob(base_directory + "/**/*.lang", recursive=True):
            local_path = os.path.relpath(local_path, self.input_path)
            self._language_files.append(LanguageFile(filepath = local_path, pack = self))
            
        return self._language_files

class Project():
    """
    Project is a class which represents an entire 'addon', via references to a 
    ResourcePack and BehaviorPack, along with some helper methods.
    """
    def __init__(self, behavior_path: str, resource_path: str) -> Project:
        self.__behavior_path = behavior_path
        self.__resource_path = resource_path
        self.__resource_pack : ResourcePack = None
        self.__behavior_pack : BehaviorPack = None
    
    def set_output_directory(self, save_location: str) -> None:
        """
        Sets the save location of the RP and the BP, based on the folder
        name from their input path.

        In other words, pass in a folder where you want both the RP and the BP 
        to be saved.

        If you need finer control, use the 'set_save_location' method on the 
        ResourcePack and BehaviorPack instead.
        """
        self.resource_pack.output_directory = save_location + "/" + os.path.dirname(self.resource_pack.input_path)
        self.behavior_pack.output_directory = save_location + "/" + os.path.dirname(self.behavior_pack.input_path)

    def get_packs(self) -> Tuple[behavior_pack, resource_pack]:
        """
        Returns Behavior Pack followed by ResourcePack.
        Useful for quickly defining rp and bp:
        rp, bp = Project("...").get_packs()
        """
        return self.behavior_pack, self.resource_pack
        
    @cached_property
    def resource_pack(self) -> ResourcePack:
        """
        The resource pack of the project.
        """
        self.__resource_pack = ResourcePack(self.__resource_path, project=self)
        return self.__resource_pack

    @cached_property
    def behavior_pack(self) -> BehaviorPack:
        """
        The behavior pack of the project.
        """
        self.__behavior_pack = BehaviorPack(self.__behavior_path, project=self)
        return self.__behavior_pack

    def save(self, force=False):
        """
        Saves both packs.
        """
        self.__behavior_pack.save(force=force)
        self.__resource_pack.save(force=force)


class Resource():
    """
    The top resource in the inheritance chain.

    Contains:
     - reference to the Pack (could be blank, if floating resource)
     - reference to the File (could be a few links up the chain)
     - dirty status
     - abstract ability to save, including context manager support
     - list of children resources
    """
    
    jsonpath : str
    
    def __init__(self, file: FileResource = None, pack: Pack = None) -> None:
        # Public
        self.pack: Union[ResourcePack, BehaviorPack] = pack
        self.file = file
        self._dirty = False
        self._deleted = False

        # Private
        self._resources: Resource = []

    def __enter__(self) -> Resource:
        """
        Context manager support.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        Context manager support. Will save the resource on exit.
        """
        self.save()
    
    @property
    def dirty(self):
        """
        Whether the asset is dirty (has been edited).
        """
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        self._dirty = dirty

    def register_resource(self, resource) -> None:
        """
        Register a child resource. These resources will always be saved first.
        """
        self._resources.append(resource)

    def _save(self):
        """
        Internal implementation of asset saving.

        Should always be implemented.
        """
        raise NotImplementedError()

    def _delete(self):
        """
        Internal implementation of asset deletion.

        Should always be implemented.
        """
        raise NotImplementedError()


    def save(self, force=False):
        """
        Saves the resource. Should not be overridden.

        This method will save all child resources first, by calling
        the internal _save implementation.
        """

        # File Assets without packs may not save.
        if isinstance(self, FileResource) and not self.pack:
            raise FloatingAssetError("Assets without a pack cannot be saved.")

        # Only dirty assets can be saved, unless forced.
        if self.dirty or force:
            # Save all children first
            for resource in self._resources:
                resource.save(force=force)

            # Now, save this resource
            self._save()

            # Mark as clean
            self.dirty = False


    def delete(self):
        """
        Deletes the resource. Should not be overridden.
        """

        # First, delete all resources of children
        for resource in self._resources:
            resource.delete()

        # Then delete self
        self._delete()


class FileResource(Resource):
    """
    A resource, which is also a file.
    Contains:
     - File path
     - Ability to mark for deletion
    """
    def __init__(self, filepath: str = None, pack: Pack = None) -> None:
        # Initialize resource
        super().__init__(file=self, pack=pack)

        # All files must register themselves with their pack
        # if the pack exists.
        if self.pack:
            pack.register_resource(self)

        # Public
        self.filepath = filepath

        if filepath:
            self.file_name = os.path.basename(filepath)

    def _delete(self):
        """
        Internal implementation of file deletion.

        Marking for deletion allows us to delete the file during saving,
        without effecting the file system during normal operation.
        """
        self._deleted = True

class JsonResource(Resource):
    """
    Parent class, which is responsible for all resources which contain
    json data.
    Should not be used directly. Use JsonFileResource, or JsonSubResource instead.
    Contains:
     - Data object
     - Method for interacting with the data
    """

    # The type information, used for generating this class at runtime.
    type_info : TypeInfo
    
    def __init__(self, data: dict = None, file: FileResource = None, pack: Pack = None) -> None:
        super().__init__(file=file, pack=pack)
        self._data = data
    
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self.dirty = True
        self._data = data

    def _save(self):
        raise NotImplementedError("This json resource cannot be saved.")

    def _delete(self):
        raise NotImplementedError("This json resource cannot be deleted.")

    def __str__(self):
        return json.dumps(self.data, indent=2, ensure_ascii=False)

    def jsonpath_exists(self, json_path:str) -> bool:
        """
        Checks if a jsonpath exists
        """
        try:
            self.get_jsonpath(json_path)
            return True
        except AssetNotFoundError:
            return False

    def delete_jsonpath(self, json_path:str) -> None:
        """
        Removes value at jsonpath location.
        """
        path_exists = self.jsonpath_exists(json_path)
        if path_exists:
            self.dirty = True
            dpath.delete(self.data, json_path)

    def pop_jsonpath(self, json_path, default=NO_ARGUMENT) \
        -> Union[dict, list, int, str, float]:
        """
        Removes value at jsonpath location, and returns it.
        """

        data = self.get_jsonpath(json_path, default=default)
        self.delete_jsonpath(json_path)
        self.dirty = True
        return data

    def append_jsonpath(self, json_path:str, insert_value:Any):
        """
        Appends a value at jsonpath location. Will create path if it doesn't exist.
        """

        path_exists = self.jsonpath_exists(json_path)

        # Otherwise, set the value
        self.dirty = True

        if path_exists:
            self.get_jsonpath(json_path).append(insert_value)
        else:
            self.set_jsonpath(json_path, [insert_value])

    def set_jsonpath(self, json_path:str, insert_value:any, overwrite:bool=True):
        """
        Sets value at jsonpath location.

        Can create a new key if it doesn't exist.
        """

        path_exists = self.jsonpath_exists(json_path)

        # If overwrite is false, it will set the path only
        # if there is no path.
        if path_exists and not overwrite:
            return

        # Otherwise, set the value
        self.dirty = True
        dpath.new(self.data, json_path, insert_value)
        

    def get_jsonpath(self, json_path, default=NO_ARGUMENT):
        """
        Gets value at jsonpath location.

        A default value may be provided, for missing keys.

        raises:
            AssetNotFoundError if the path does not exist.
        """
        try:
            return dpath.get(self.data, json_path)
        except Exception as exception:
            if default is not NO_ARGUMENT:
                return default
            raise AssetNotFoundError(
                f"Path {json_path} does not exist."
            ) from exception

    def get_data_at(self, json_path):
        """
        Returns a list of jsonpaths found at this jsonpath location.
        For a list, this will return: json_path + list index
        For a dict, this will return json_path + dict key

        Exception will be raised if the result is not a list or dict.
        Missing data path will return []
        """
        try:
            # Special case, to handle the root of the json
            if json_path == "**":
                result = self.data
            else:
                result = self.get_jsonpath(json_path, default=[])

            if isinstance(result, dict):
                for key in result.keys():
                    yield json_path + f"/{key}", result[key]
            elif isinstance(result, list):
                for i, element in enumerate(result):
                    yield json_path + f"/[{i}]", element
            else:
                raise AmbiguousAssetError(f"Path '{json_path}' matched a single element, not a list or dict.")

        except KeyError as key_error:
            raise AssetNotFoundError(json_path, self.data) from key_error


class JsonSubResource(JsonResource):
    """
    A sub resource represents a chunk of json data, within a file.
    """
    def __init__(self, parent: Resource = None, json_path: str = None, data: dict = None) -> None:
        super().__init__(data = data, pack = parent.pack, file = parent.file)

        # The parent of the sub-resource is the asset of type Resource
        # which owns this sub-resource. For example a Component is owned by
        # either an EntityFileBP, or a ComponentGroup.
        self.parent: JsonResource = parent

        # The jsonpath is the location within the parent resource, where
        # this sub-resource is stored.
        self._json_path: str = json_path

        # The original path location needs to be stored, for the purpose
        # of renaming or moving the sub-resource.
        self._original_json_path: str = json_path

        # Register self into parent, so that it can be found by the parent
        # during saving, etc.
        self.parent.register_resource(self)
    
    @property
    def json_path(self):
        return self._json_path

    @json_path.setter
    def json_path(self, json_path):
        self.dirty = True
        self._json_path = json_path

    @property
    def id(self):
        """
        The ID of the sub-resource, such as 'minecraft:scale' for a component.
        """
        return self.json_path.rsplit("/", maxsplit=1)[1]

    @id.setter
    def id(self, id):
        self.dirty = True
        self.json_path = self.json_path.rsplit("/", maxsplit=1)[0] + "/" + id

    def __repr__(self):
        return f"'{self.__class__.__name__}: {self.id}'"

    def __str__(self):
        """
        Returns string representation of this resource.

        The json data will be representation for printing: With the id appended,
        and with indentation.
        """
        return f'"{self.id}": {json.dumps(self.data, indent=2, ensure_ascii=False)}'

    @property
    def dirty(self):
        """
        Returns True if this resource has been modified since it was last saved.
        """
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        """
        Set the dirty flag, which stores whether this resource has been
        modified since it was last saved.
        """
        self._dirty = dirty
        self.parent.dirty = dirty # Pseudo recursive call will travel up the chain


    def _save(self):
        """
        Saves the resource, into its parents json structure.

        This works by replacing the data at the jsonpath location,
        meaning that the parent will contain accurate representation of
        the children's data, saved into itself.
        """

        # If the id was updated, we need to serialize into a new location.
        # This implies first deleting the old location, and then creating
        # a new one.
        if not self._deleted:
            if self.json_path != self._original_json_path:
                self.parent.delete_jsonpath(self._original_json_path)
                self.parent.set_jsonpath(self.json_path, self.data)
            else:
                self.parent.set_jsonpath(self.json_path, self.data)

        self._dirty = False

    def _delete(self):
        """
        Deletes itself from parent, by removing the data at the jsonpath
        location.

        If it is a list, it must only set itself to None. The actual
        deletion will take place during saving.
        """
        self.parent.dirty = True
        self.parent._resources.remove(self)
        self._deleted = True
        self.parent.delete_jsonpath(self.json_path)

class JsonFileResource(FileResource, JsonResource):
    """
    A file, which contains json data. Most files in the addon system
    are of this type, or have it as a resource parent.
    """
    def __init__(self, data: dict = None, filepath: str = None, pack: Pack = None) -> None:
        # Init file resource parent, which gives us access to data
        FileResource.__init__(self, filepath=filepath, pack=pack)
        
        # Data is either set directly, or is read from the filepath for this
        # resource. This allows assets to be created from scratch, whilst
        # still having an associated file location.
        if data is not None:
            self.data = data
        else:
            self.data = self.load_json(self.filepath)

        # Init json resource, which relies on the new data attribute
        JsonResource.__init__(self, data=self.data, file=self, pack=pack)

    def __repr__(self):
        return f"'{self.__class__.__name__}: {self.filepath}'"


    def load_json(self, filepath: str) -> dict:
        """
        Loads json from file. `local_path` paramater is local to the projects
        input path.
        """
        
        if self.pack:
            filepath = os.path.join(self.pack.input_path, filepath)

        if not os.path.exists(filepath):
            raise AssetNotFoundError(f"File not found: {filepath}")
        try:
            with open(filepath, "r", encoding='utf8') as fh:
                try:
                    return json.load(fh)
                except json.JSONDecodeError:
                    try:
                        fh.seek(0)
                        contents = ""
                        for line in fh.readlines():
                            cleaned_line = line.split("//", 1)[0]
                            if len(cleaned_line) > 0 and line.endswith("\n") and "\n" not in cleaned_line:
                                cleaned_line += "\n"
                            contents += cleaned_line
                        while "/*" in contents:
                            pre_comment, post_comment = contents.split("/*", 1)
                            contents = pre_comment + post_comment.split("*/", 1)[1]
                        return json.loads(contents)
                    except json.JSONDecodeError as exception:
                        return {}
        except Exception:
            raise InvalidJsonError(filepath)

    def _save(self):
        save_path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(save_path)

        # If the file has been marked for deletion, delete it
        if self._deleted:
            # If the paths are the same, delete the file, otherwise
            # we can just pass
            if smart_compare(self.pack.input_path, self.pack.output_directory):
                os.remove(save_path)
        else:
            with open(save_path, "w+") as file_head:
                clean_data = {k: v for k, v in self.data.items() if v is not None}
                return json.dump(clean_data, file_head, indent=2, ensure_ascii=False)

class LanguageFile(FileResource):
    """
    A LanguageFile is a language file, such as 'en_US.lang', and is made
    up of many Translations.
    """
    def __init__(self, filepath: str = None, pack: Pack = None) -> None:
        super().__init__(filepath=filepath, pack=pack)
        self.__translations: list[Translation] = []
    
    def get_translation(self, key: str) -> Translation:
        """
        Whether the language file contains the specified key.
        """
        for translation in self.translations:
            if translation.key == key:
                return translation 
        raise AssetNotFoundError(f"Translation with key '{key}' not found in language file '{self.filepath}'.")

    def contains_translation(self, key: str) -> bool:
        """
        Whether the language file contains the specified key.
        """

        for translation in self.translations:
            if translation.key == key:
                return True
        return False

    def delete_translation(self, key: str) -> None:
        """
        Deletes a translation based on key, if it exists.
        """
        for translation in self.translations:
            if translation.key == key:
                self.dirty = True
                self.translations.remove(translation)
                return

    def add_translation(self, translation: Translation, overwrite: bool = True) -> bool:
        """
        Adds a new translation key. Overwrites by default.
        """

        # We must complain about duplicates.
        # If no overwrite, don't add
        # If overwrite, delete previous translation
        if self.contains_translation(translation.key):
            if not overwrite:
                return False
            else:
                self.delete_translation(translation.key)

        self.dirty = True
        self.__translations.append(translation)

        return True

    def _save(self):
        path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(path)
        with open(path, 'w', encoding='utf-8') as file:
            for translation in self.translations:
                file.write(f"{translation.key}={translation.value}\t##{translation.comment}\n")


    @cached_property
    def translations(self) -> list[Translation]:
        with open(os.path.join(self.pack.input_path, self.filepath), "r", encoding='utf-8') as language_file:
            for line in language_file.readlines():
                language_regex = "^([^#\n]+?)=([^#]+)#*?([^#]*?)$"
                if match := re.search(language_regex, line):
                    groups = match.groups()
                    self.__translations.append(
                        Translation(
                            key = groups[0].strip() if len(groups) > 0 else "",
                            value = groups[1].strip() if len(groups) > 1 else "",
                            comment = groups[2].strip() if len(groups) > 2 else "",
                        )
                    )
        return self.__translations

# Generic Resources
class Translation():
    """
    Dataclass for a translation. Many translations together make up a
    TranslationFile.
    """

    def __init__(self, key: str, value: str, comment: str = "") -> None:
        self.key = key
        self.value = value
        self.comment = comment

class FormatVersion():
    def __init__(self, version) -> None:
        if isinstance(version, FormatVersion):
            self.major = version.major
            self.minor = version.minor
            self.patch = version.patch
            return
        elif isinstance(version, str):
            elements = version.split('.')
        else:
            # Change to suitable error
            raise TypeError()

        # Pack with extra data if it's missing
        for i in range(3 - len(elements)):
            elements.append('0')
        
        self.major = int(elements[0])
        self.minor = int(elements[1])
        self.patch = int(elements[2])

    def __repr__(self) -> str:
        return f'{self.major}.{self.minor}.{self.patch}'
        
    def __eq__(self, other):
        other = FormatVersion(other)
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __gt__(self, other):
        if self.major > other.major:
            return True
        elif self.major < other.major:
            return False

        if self.minor > other.minor:
            return True
        elif self.minor < other.minor:
            return False

        if self.patch > other.patch:
            return True
        elif self.patch < other.patch:
            return False
        
        return self != other


class AnimationControllerBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animation_controllers",
        attribute = "animation_controllers",
        getter_attribute = "id"
    )

@format_version()
class AnimationControllerFileBP(JsonFileResource):

    format_version : FormatVersion
    type_info = TypeInfo(
        filepath = "animation_controllers",
        attribute = "animation_controller_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(AnimationControllerBP)
    def animation_controllers(self): pass
    @Getter(AnimationControllerBP)
    def get_animation_controller(self, id: str): pass
    @SubResourceAdder(AnimationControllerBP)
    def add_animation_controller(self, name: str, data: dict): pass


class FunctionFile(FileResource):
    """
    A FunctionFile is a function file, such as run.mcfunction, and is
    made up of many commands.
    """
    
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._commands : list[Command] = []

    type_info = TypeInfo(
        filepath = "functions",
        extension = ".mcfunction",
        attribute = "functions",
        getter_attribute="filepath"
    )
    
    def strip_comments(self):
        """
        Strips all comments from the function file.
        Generally should be used before accessing and using the `commands` property.
        """
        self.commands = [c for c in self.commands if not c.is_comment()]

    @cached_property
    def commands(self) -> list[Command]:
        """
        The list of commands in this function file. Every line represents a Command.
        """
        with open(os.path.join(self.pack.input_path, self.filepath), "r", encoding='utf-8') as function_file:
            for line in function_file.readlines():
                command = line.strip()
                if command:
                    self._commands.append(Command(command, file=self, pack=self.pack))
        self._commands = convert_to_notify_structure(self._commands, self)
        return self._commands
    
    def _save(self) -> None:
        """
        Writes the commands back to the file, one command at a time.
        """
        path = os.path.join(self.pack.output_directory, self.filepath)
        create_nested_directory(path)
        with open(path, 'w', encoding='utf-8') as file:
            for command in self.commands:
                file.write(command.data + '\n')

@identifier("minecraft:feature_rules/description/identifier")
@format_version()
class FeatureRuleFile(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "feature_rules",
        attribute = "feature_rules",
        getter_attribute="identifier"
    )

@identifier("**/description/identifier")
@format_version()
class FeatureFile(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "features",
        attribute = "features",
        getter_attribute="identifier"
    )
    
@format_version()
@identifier("minecraft:spawn_rules/description/identifier")
class SpawnRuleFile(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "spawn_rules",
        attribute = "spawn_rules",
        getter_attribute="identifier"
    )

@identifier("**/identifier")
@format_version()
class RecipeFile(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "recipes",
        attribute = "recipes",
        getter_attribute="identifier"
    )

class EntityComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/components",
        attribute = "components",
        getter_attribute = "id"
    )

class EntityEventBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/events",
        attribute = "events",
        getter_attribute = "id"
    )

class Component(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "**",
        attribute = "components",
        getter_attribute = "id"
    )

class ComponentGroup(JsonSubResource):
    """
    A component group is a collection of components in an EntityFileBP.
    """
    type_info = TypeInfo(
        jsonpath = "minecraft:entity/component_groups",
        attribute = "component_groups",
        getter_attribute = "id"
    )

    @SubResourceDefinition(Component)
    def components(self): pass
    @Getter(Component)
    def get_component(self, id: str): pass
    @SubResourceAdder(Component)
    def add_component(self, name: str, data: dict): pass

@format_version()
@identifier("minecraft:entity/description/identifier")
class EntityFileBP(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "entities",
        attribute = "entities",
        getter_attribute = "identifier"
    )

    @property
    def counterpart(self) -> EntityFileRP:
        return self.pack.project.resource_pack.get_entity(self.identifier)

    @SubResourceDefinition(EntityEventBP)
    def events(self): pass
    @Getter(EntityEventBP)
    def get_event(self, id: str): pass
    @SubResourceAdder(EntityEventBP)
    def add_event(self, name: str, data: dict): pass

    @SubResourceDefinition(EntityComponentBP)
    def components(self): pass
    @Getter(EntityComponentBP)
    def get_component(self, id: str): pass
    @SubResourceAdder(EntityComponentBP)
    def add_component(self, name: str, data: dict): pass

    @SubResourceDefinition(ComponentGroup)
    def component_groups(self): pass
    @Getter(ComponentGroup)
    def get_component_group(self, id: str): pass
    @SubResourceAdder(ComponentGroup)
    def add_component_group(self, name: str, data: dict): pass

class LootTablePool(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "pools",
        attribute = "pools"
    )

class LootTableFile(JsonFileResource):
    type_info = TypeInfo(
        filepath = "loot_tables",
        attribute = "loot_tables",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(LootTablePool)
    def pools(self): pass

class ItemComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/components",
        attribute = "components",
        getter_attribute = "id"
    )

class ItemEventBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/events",
        attribute = "events",
        getter_attribute = "id"
    )


@format_version()
@identifier("minecraft:item/description/identifier")
class ItemFileBP(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "items",
        attribute = "items",
        getter_attribute = "identifier"
    )

    @SubResourceDefinition(ItemComponentBP)
    def components(self): pass
    @Getter(ItemComponentBP)
    def get_component(self, id: str): pass
    @SubResourceAdder(ItemComponentBP)
    def add_component(self, name: str, data: dict): pass

    @SubResourceDefinition(ItemEventBP)
    def events(self): pass
    @Getter(ItemEventBP)
    def get_event(self, id: str): pass
    @SubResourceAdder(ItemEventBP)
    def add_event(self, name: str, data: dict): pass


class BlockFileComponentBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:block/components",
        attribute = "components",
        getter_attribute = "id"
    )


@format_version()
@identifier(jsonpath="minecraft:block/description/identifier")
class BlockFileBP(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "blocks",
        attribute = "blocks",
        getter_attribute = "identifier"
    )

    @SubResourceDefinition(BlockFileComponentBP)
    def components(self): pass
    @Getter(BlockFileComponentBP)
    def get_component(self, id: str): pass
    @SubResourceAdder(BlockFileComponentBP)
    def add_component(self, name: str, data: dict): pass

@ClassProperty('loop')
class AnimationBP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animations",
        attribute = "animations",
        getter_attribute = "id"
    )

class AnimationFileBP(JsonFileResource):
    type_info = TypeInfo(
        filepath = "animations",
        attribute = "animation_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(AnimationBP)
    def animations(self): pass
    @SubResourceAdder(AnimationBP)
    def add_animation(self, name: str, data: dict): pass
    @Getter(AnimationBP)
    def get_animation(self, id: str): pass

class BehaviorPack(Pack):
    """
    The BehaviorPack represents the behavior pack of a project.
    """

    @ResourceDefinition(AnimationControllerFileBP)
    def animation_controller_files(self): pass
    @Getter(AnimationControllerFileBP)
    def get_animation_controller_file(self, filepath: str): pass
    @JsonChildResource(AnimationControllerFileBP, AnimationControllerBP)
    def animation_controllers(self): pass
    @ChildGetter(AnimationControllerFileBP, AnimationControllerBP)
    def get_animation_controller(self, id: str): pass

    @ResourceDefinition(AnimationFileBP)
    def animation_files(self): pass
    @Getter(AnimationFileBP)
    def get_animation_file(self, filepath: str): pass
    @JsonChildResource(AnimationFileBP, AnimationBP)
    def animations(self): pass
    @ChildGetter(AnimationFileBP, AnimationBP)
    def get_animation(self, id: str): pass

    @ResourceDefinition(FunctionFile)
    def functions(self): pass
    @Getter(FunctionFile)
    def get_function(self, filepath: str): pass

    @ResourceDefinition(FeatureRuleFile)
    def feature_rules(self): pass
    @Getter(FeatureRuleFile)
    def get_feature_rule(self, identifier: str): pass

    @ResourceDefinition(FeatureFile)
    def features(self): pass
    @Getter(FeatureFile)
    def get_feature(self, identifier: str): pass

    @ResourceDefinition(SpawnRuleFile)
    def spawn_rules(self): pass
    @Getter(SpawnRuleFile)
    def get_spawn_rule(self, identifier: str): pass

    @ResourceDefinition(RecipeFile)
    def recipes(self): pass
    @Getter(RecipeFile)
    def get_recipe(self, identifier: str): pass

    @ResourceDefinition(EntityFileBP)
    def entities(self): pass
    @Getter(EntityFileBP)
    def get_entity(self, identifier: str): pass

    @ResourceDefinition(LootTableFile)
    def loot_tables(self): pass
    @Getter(LootTableFile)
    def get_loot_table(self, identifier: str): pass

    @ResourceDefinition(ItemFileBP)
    def items(self): pass
    @Getter(ItemFileBP)
    def get_item(self, identifier: str): pass

    @ResourceDefinition(BlockFileBP)
    def blocks(self): pass
    @Getter(BlockFileBP)
    def get_block(self, identifier: str): pass


class Event(JsonSubResource):
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)
        
        self.__groups_to_add = []
        self.__groups_to_remove = []

    @cached_property
    def groups_to_add(self) -> list[ComponentGroup]:
        for path, data in self.get_data_at("add/component_groups"):
            self.__groups_to_add.append(ComponentGroup(parent = self, json_path = path, data = data))
        return self.__groups_to_add
    
    @cached_property
    def groups_to_remove(self) -> list[ComponentGroup]:
        for path, data in self.get_data_at("remove/component_groups"):
            self.__groups_to_remove.append(ComponentGroup(parent = self, json_path = path, data = data))
        return self.__groups_to_remove

class Command(Resource):
    """
    A command is a wrapper around a string, which represents a single command
    in a function file.

    To use this class, you can access the 'data' property, and treat it like
    a string.
    """
    def __init__(self, command: str, file: FileResource = None, pack: Pack = None) -> None:
        super().__init__(file=file, pack=pack)

        # The 'data' is the actual command, which is stored as a string.
        self._data: str = command

    @property
    def data(self):
        return self._data
    
    @data.setter
    def data(self, data):
        self._data = data
        self.dirty = True

    @property
    def dirty(self):
        return self._dirty

    @dirty.setter
    def dirty(self, dirty):
        """
        When a command is marked as dirty, it must propagate this to the
        file, so that the file can be marked as dirty.
        """
        self._dirty = dirty
        self.file.dirty = dirty

    def is_comment(self):
        return self.data.startswith("#")

    def __str__(self):
        return self.data

    def __repr__(self):
        return f"Command: '{self.data}'"

class ParticleFileComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "particle_effect/components",
        attribute = "components",
        getter_attribute = "id"
    )

class ParticleFileEvent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "particle_effect/events",
        attribute = "components",
        getter_attribute = "id"
    )

@format_version()
@identifier("particle_effect/description/identifier")
class ParticleFile(JsonFileResource):
    """
    ParticleFile is a JsonFileResource which represents a particle file.
    """
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "particles",
        attribute = "particles",
        getter_attribute = "identifier"
    )

    @SubResourceDefinition(ParticleFileComponent)
    def components(self): pass
    @Getter(ParticleFileComponent)
    def get_component(self, id: str): pass
    @SubResourceAdder(ParticleFileComponent)
    def add_component(self, name: str, data: dict): pass

    @SubResourceDefinition(ParticleFileEvent)
    def events(self): pass
    @Getter(ParticleFileEvent)
    def get_event(self, id: str): pass
    @SubResourceAdder(ParticleFileEvent)
    def add_event(self, name: str, data: dict): pass

class AnimationControllerStateRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "states",
        attribute = "states",
        getter_attribute = "id"
    )

@ClassProperty('initial_state')
class AnimationControllerRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animation_controllers",
        attribute = "animation_controllers",
        getter_attribute = "id"
    )

    @SubResourceDefinition(AnimationControllerStateRP)
    def states(self): pass
    @Getter(AnimationControllerStateRP)
    def get_state(self, id: str): pass
    @SubResourceAdder(AnimationControllerStateRP)
    def add_state(self, name: str, data: dict): pass

class AnimationControllerFileRP(JsonFileResource):
    """
    AnimationControllerFileRP
    """
    type_info = TypeInfo(
        filepath = "animation_controllers",
        attribute = "animation_controller_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(AnimationControllerRP)
    def animation_controllers(self): pass
    @Getter(AnimationControllerRP)
    def get_animation_controller(self, id: str): pass
    @SubResourceAdder(AnimationControllerRP)
    def add_animation_controller(self, name: str, data: dict): pass

@ClassProperty('loop')
class AnimationRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "animations",
        attribute = "animations",
        getter_attribute = "id"
    )

class AnimationFileRP(JsonFileResource):
    """
    AnimationFileRP is a class which represents a resource pack's animation file.
    Since many animations are defined in the same file, it is often more useful.
    to use the AnimationRP class instead.
    """

    type_info = TypeInfo(
        filepath = "animations",
        attribute = "animation_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(AnimationRP)
    def animations(self): pass
    @SubResourceAdder(AnimationRP)
    def add_animation(self, name: str, data: dict): pass
    @Getter(AnimationRP)
    def get_animation(self, id: str): pass

@format_version()
@identifier("minecraft:attachable/description/identifier")
class AttachableFileRP(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "attachables",
        attribute = "attachables",
        getter_attribute = "identifier"
    )


class BiomesClientFile(JsonFileResource):
    """
    BiomesClientFile is a class which represents the data stored in 'rp/biomes_client.json'
    """

    type_info = TypeInfo(
        filepath = "biomes_client",
        attribute = "biomes_client_file"
    )


class BlocksFile(JsonFileResource):
    """
    BlocksFile is a class which represents the data stored in 'rp/blocks.json'
    """

    type_info = TypeInfo(
        filepath = "blocks",
        attribute = "blocks_file"
    )



@format_version()
@identifier("minecraft:client_entity/description/identifier")
class EntityFileRP(JsonFileResource):
    """
    EntityFileRP is a class which represents a resource pack's entity file.
    """
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "entity",
        attribute = "entities",
        getter_attribute = "identifier"
    )

    @property
    def counterpart(self) -> EntityFileBP:
        return self.pack.project.behavior_pack.get_entity(self.identifier)

    @SubResourceDefinition(EntityEventBP)
    def events(self): pass
    @Getter(EntityEventBP)
    def get_event(self, id: str): pass
    @SubResourceAdder(EntityEventBP)
    def add_event(self, name: str, data: dict): pass

    # TODO: See if the decoration system can apply to these 'triples'.
    # Or we need to add 'adders' for all three (anim, texture, model)
    @cached_property
    def animations(self) -> list[AnimationTriple]:
        self._animations = []
        for path, data in self.get_data_at("minecraft:client_entity/description/animations"):
            self._animations.append(AnimationTriple(parent = self, json_path = path, data = data))
        return self._animations
    
    def get_animation(self, identifier:str) -> AnimationTriple:
        """
        Fetches an AnimationTriple resource, either by shortname, or identifier.
        """
        for child in self.animations:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def textures(self) -> list[TextureDouble]:
        self._textures = []
        for path, data in self.get_data_at("minecraft:client_entity/description/textures"):
            self._textures.append(TextureDouble(parent = self, json_path = path, data = data))
        return self._textures
    
    def get_texture(self, identifier:str) -> TextureDouble:
        """
        Fetches a texture resource, either by shortname, or texture_path.
        """
        for child in self.textures:
            if smart_compare(child.shortname, identifier) or smart_compare(child.texture_path, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def models(self) -> list[ModelTriple]:
        self._models = []
        for path, data in self.get_data_at("minecraft:client_entity/description/geometry"):
            self._models.append(ModelTriple(parent = self, json_path = path, data = data))
        return self._models
    
    def get_model(self, identifier:str) -> ModelTriple:
        """
        Fetches a model resource, either by shortname, or identifier.
        """
        for child in self.models:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)

    @cached_property
    def materials(self) -> list[MaterialTriple]:
        self._materials = []
        for path, data in self.get_data_at("minecraft:client_entity/description/materials"):
            self._materials.append(MaterialTriple(parent = self, json_path = path, data = data))
        return self._materials
    
    def get_material(self, identifier:str) -> MaterialTriple:
        """
        Fetches a material resource, either by shortname, or material type.
        """
        for child in self.materials:
            if smart_compare(child.shortname, identifier) or smart_compare(child.identifier, identifier):
                return child
        raise AssetNotFoundError(id)


class FlipbookTexturesFile(JsonFileResource):
    """
    FlipbookTexturesFile is a class which represents the data stored in 'rp/textures/flipbook_textures.json'
    """

    type_info = TypeInfo(
        filepath = "textures/flipbook_textures",
        attribute = "flipbook_textures_file"
    )


class FogDistanceComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/distance",
        attribute = "distance_components",
        getter_attribute = "id"
    )

class FogVolumetricDensityComponent(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/volumetric/density",
        attribute = "volumetric_density_components",
        getter_attribute = "id"
    )

class FogVolumetricMediaCoefficient(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:fog_settings/volumetric/media_coefficients",
        attribute = "volumetric_media_coefficients",
        getter_attribute = "id"
    )

@format_version()
@identifier("minecraft:fog_settings/description/identifier")
class FogFile(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "fogs",
        attribute = "fogs",
        getter_attribute = "identifier"
    )

    @SubResourceDefinition(FogDistanceComponent)
    def distance_components(self): pass
    @Getter(FogDistanceComponent)
    def get_distance_component(self, id: str): pass
    @SubResourceAdder(FogDistanceComponent)
    def add_distance_component(self, name: str, data: dict): pass

    @SubResourceDefinition(FogVolumetricDensityComponent)
    def volumetric_density_components(self): pass
    @Getter(FogVolumetricDensityComponent)
    def get_volumetric_density_component(self, id: str): pass
    @SubResourceAdder(FogVolumetricDensityComponent)
    def add_volumetric_density_component(self, name: str, data: dict): pass

    @SubResourceDefinition(FogVolumetricMediaCoefficient)
    def volumetric_media_coefficients(self): pass
    @Getter(FogVolumetricMediaCoefficient)
    def get_volumetric_media_coefficient(self, id: str): pass
    @SubResourceAdder(FogVolumetricMediaCoefficient)
    def add_volumetric_media_coefficient(self, name: str, data: dict): pass

class ItemComponentRP(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "minecraft:item/components",
        attribute = "components",
        getter_attribute = "id"
    )

@format_version()
@identifier("minecraft:item/description/identifier")
class ItemFileRP(JsonFileResource):
    format_version : FormatVersion
    identifier: str
    type_info = TypeInfo(
        filepath = "items",
        attribute = "items",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(ItemComponentRP)
    def components(self): pass
    @Getter(ItemComponentRP)
    def get_component(self, id: str): pass
    @SubResourceAdder(ItemComponentRP)
    def add_component(self, name: str, data: dict): pass

class Material(JsonSubResource):
    """
    Represents a single material, from a .material file
    """

    type_info = TypeInfo(
        jsonpath = "**",
        attribute = "materials",
        getter_attribute = "id"
    )

@format_version(jsonpath="materials/version")
class MaterialFile(JsonFileResource):
    """
    MaterialFile is a class which represents a resource pack's material file.
    Since many materials can be defined in the same file, it is often more useful
    to use the MaterialRP class directly.
    """
    format_version : FormatVersion
    type_info = TypeInfo(
        filepath = "materials",
        attribute = "material_files",
        getter_attribute = "filepath",
        extension="material"
    )

    @SubResourceDefinition(Material)
    def materials(self): pass
    @Getter(Material)
    def get_material(self, id: str): pass
    @SubResourceAdder(Material)
    def add_material(self, name: str, data: dict): pass

class Cube(JsonSubResource):
    type_info = TypeInfo(
        jsonpath = "cubes",
        attribute = "cubes",
        getter_attribute = "id"
    )

@ClassProperty("name")
class Bone(JsonSubResource):
    name: str
    type_info = TypeInfo(
        jsonpath = "bones",
        attribute = "bones",
        getter_attribute = "name"
    )

    @SubResourceDefinition(Cube)
    def cubes(self): pass
    @Getter(Cube)
    def get_cube(self, id: str): pass
    @SubResourceAdder(Cube)
    def add_cube(self, name: str, data: dict): pass


@identifier("description/identifier")
class Model(JsonSubResource):
    identifier: str
    type_info = TypeInfo(
        jsonpath = "minecraft:geometry",
        attribute = "models",
        getter_attribute = "identifier"
    )

    @SubResourceDefinition(Bone)
    def bones(self): pass
    @Getter(Bone)
    def get_bone(self, id: str): pass
    @SubResourceAdder(Bone)
    def add_bone(self, name: str, data: dict): pass

@format_version()
class ModelFile(JsonFileResource):
    format_version : FormatVersion
    type_info = TypeInfo(
        filepath = "models",
        attribute = "model_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(Model)
    def models(self): pass
    @Getter(Model)
    def get_model(self, id: str): pass
    @SubResourceAdder(Model)
    def add_model(self, name: str, data: dict): pass

class RenderController(JsonSubResource):
    """
    A JsonSubResource, representing a single Render Controller object, contained
    within a RenderControllerFile.
    """

    type_info = TypeInfo(
        jsonpath = "render_controllers",
        attribute = "render_controllers",
        getter_attribute = "id"
    )

@format_version()
class RenderControllerFile(JsonFileResource):
    format_version : FormatVersion
    type_info = TypeInfo(
        filepath = "render_controllers",
        attribute = "render_controller_files",
        getter_attribute = "filepath"
    )

    @SubResourceDefinition(RenderController)
    def render_controllers(self): pass
    @SubResourceAdder(RenderController)
    def add_render_controller(self, name: str, data: dict): pass
    @Getter(RenderController)
    def get_render_controller(self, id: str): pass


@format_version()
class SoundDefinitionsFile(JsonFileResource):
    """
    SoundsDefinitionFile is a class which represents the data stored in
    'rp/sounds/sound_definitions.json'
    """
    format_version : FormatVersion
    type_info = TypeInfo(
        filepath = "sounds/sound_definitions",
        attribute = "sound_definitions_file"
    )


class SoundsFile(JsonFileResource):
    """
    SoundsFile is a class which represents the data stored in 'rp/sounds.json'
    """
    type_info = TypeInfo(
        filepath = "sounds",
        attribute = "sounds_file"
    )

class StandAloneTextureFile(JsonFileResource):
    """
    StandAloneTextureFile is a class which represents the data stored in 'rp/textures/*_texture.json'.
    Examples: 'item_texture.json', 'terrain_texture.json.

    These actual children are subclassed
    """
    def __init__(self, data: dict = None, filepath: str = None, pack: Pack = None) -> None:
        super().__init__(data = data, filepath = filepath, pack = pack)
        self.__texture_definitions = []
    
    @cached_property
    def texture_definitions(self) -> list[TextureFileDouble]:
        for path, data in self.get_data_at("texture_data"):
            self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = path, data = data))
        return self.__texture_definitions

    def get_texture_definition(self, shortname: str) -> TextureFileDouble:
        for child in self.texture_definitions:
            if child.shortname == shortname:
                return child
        raise AssetNotFoundError(f"Texture definition for shortname '{shortname}' not found.")

    def add_texture_definition(self, shortname: str, textures: list[str]):
        self.set_jsonpath(f"texture_data/{shortname}", {
            "textures": textures
        })
        self.__texture_definitions.append(TextureFileDouble(parent = self, json_path = f"texture_data/{shortname}", data = {"textures": textures}))

class TerrainTextureFile(StandAloneTextureFile):
    type_info = TypeInfo(
        filepath = "textures/terrain_texture",
        attribute = "terrain_texture_file"
    )

class ItemTextureFile(StandAloneTextureFile):
    type_info = TypeInfo(
        filepath = "textures/item_texture",
        attribute = "item_texture_file"
    )
    

class ResourceTriple(JsonSubResource):
    """
    Base class for handling "shortname": "identifier" pairs, with an underlying, resource.
    """
    @property
    def shortname(self):
        """
        This represents the shortname of the resource. e.g., "shortname": "identifier"
        """
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def identifier(self):
        """
        This represents the identifier of the resource. e.g., "shortname": "identifier"
        """
        return self.data
    
    @identifier.setter
    def identifier(self, identifier):
        self.data = identifier

    def resource():
        """
        Returns the identifier associated with the resource.
        """
        raise NotImplementedError

class MaterialTriple(ResourceTriple):
    """
    A special sub-resource, which represents a material within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_material(self.identifier)


class AnimationTriple(ResourceTriple):
    """
    A special sub-resource, which represents an animation within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_animation(self.identifier)
    
class ModelTriple(ResourceTriple):
    """
    A special sub-resource, which represents a model within an RP entity.
    """

    @property
    def resource(self):
        return self.parent.pack.get_model(self.identifier)
    

class TextureDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within an RP entity.
    """
    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)

    @property
    def shortname(self):
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def texture_path(self):
        return self.data
    
    @texture_path.setter
    def texture_path(self, texture_path):
        self.data = texture_path
    
    def exists(self) -> bool:
        """
        Returns True if this resource exists in the pack.
        """
        return os.path.exists(os.path.join(self.pack.input_path, self.filepath))


class TextureFileDouble(JsonSubResource):
    """
    A special sub-resource, which represents a texture within a texture
    definition file, such as 'item_texture.json'

    This is a special case, as it has some additional logic for obscuring the
    texture path. This is because textures are stored nested under 
    'textures' which is simply inconvenient to work with.
    """

    def __init__(self, data: dict = None, parent: Resource = None, json_path: str = None ) -> None:
        super().__init__(data=data, parent=parent, json_path=json_path)
 
        self.textures = data.get("textures", [])
        if not isinstance(self.textures, list):
            self.textures = [self.textures]
        
        self.textures = convert_to_notify_structure(self.textures, self)
        
    @property
    def shortname(self):
        return self.id

    @shortname.setter
    def shortname(self, shortname):
        self.id = shortname
    
    @property
    def data(self):
        """
        Custom data getter allows us to re-create the json structure based
        on the saved textures.
        """
        return {
            "textures": self.textures
        }

class ResourcePack(Pack):
    def __init__(self, input_path: str, project: Project = None):
        super().__init__(input_path, project=project)
        self._sounds: list[str] = []
        self._textures: list[str] = []

    @ResourceDefinition(ParticleFile)
    def particles(self): pass
    @Getter(ParticleFile)
    def get_particle(self, identifier: str): pass

    @ResourceDefinition(AttachableFileRP)
    def attachables(self): pass
    @Getter(AttachableFileRP)
    def get_attachable(self, identifier:str): pass

    @ResourceDefinition(AnimationControllerFileRP)
    def animation_controller_files(self): pass
    @Getter(AnimationControllerFileRP)
    def get_animation_controller_file(self, filepath:str): pass
    @JsonChildResource(AnimationControllerFileRP, AnimationControllerRP)
    def animation_controllers(self): pass
    @ChildGetter(AnimationControllerFileRP, AnimationControllerRP)
    def get_animation_controller(self, id: str): pass

    @ResourceDefinition(AnimationFileRP)
    def animation_files(self): pass
    @Getter(AnimationFileRP)
    def get_animation_file(self, filepath:str): pass
    @JsonChildResource(AnimationFileRP, AnimationRP)
    def animations(self): pass
    @ChildGetter(AnimationFileRP, AnimationRP)
    def get_animation(self, id: str): pass

    @ResourceDefinition(MaterialFile)
    def material_files(self): pass
    @Getter(MaterialFile)
    def get_material_file(self, filepath:str): pass
    @JsonChildResource(MaterialFile, Material)
    def materials(self): pass
    @ChildGetter(MaterialFile, Material)
    def get_material(self, id: str): pass

    @ResourceDefinition(EntityFileRP)
    def entities(self): pass
    @Getter(EntityFileRP)
    def get_entity(self, identifier:str): pass

    @ResourceDefinition(FogFile)
    def fogs(self): pass
    @Getter(FogFile)
    def get_fog(self, identifier:str): pass

    @ResourceDefinition(ModelFile)
    def model_files(self): pass
    @Getter(ModelFile)
    def get_model_file(self, filepath:str): pass
    @JsonChildResource(ModelFile, Model)
    def models(self): pass
    @ChildGetter(ModelFile, Model)
    def get_model(self, id: str): pass

    @ResourceDefinition(RenderControllerFile)
    def render_controller_files(self): pass
    @Getter(RenderControllerFile)
    def get_render_controller_file(self, filepath:str): pass
    @JsonChildResource(RenderControllerFile, RenderController)
    def render_controllers(self): pass
    @ChildGetter(RenderControllerFile, RenderController)
    def get_render_controller(self, id: str): pass

    @ResourceDefinition(ItemFileRP)
    def items(self): pass
    @Getter(ItemFileRP)
    def get_item(self, identifier:str): pass

    @cached_property
    def sounds(self) -> list[str]:
        """
        Returns a list of all sounds in the pack, relative to the pack root.
        """
        glob_pattern = os.path.join(self.input_path, "sounds") + "/**/*."
        for extension in SOUND_EXTENSIONS:
            self._sounds.extend(glob.glob(glob_pattern + extension, recursive=True))

        self._sounds = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in self._sounds]
        return self._sounds

    def get_sounds(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child sounds of the searchpath, relative to the pack root. 
        Search path should not include 'sounds'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_sounds("entities", trim_extension=True)
        """
        sounds = []
        glob_pattern = os.path.join(self.input_path, "sounds", search_path) + "/**/*."
        for extension in SOUND_EXTENSIONS:
            sounds.extend(glob.glob(glob_pattern + extension, recursive=True))

        sounds = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in sounds]
        if trim_extension:
            sounds = [os.path.splitext(path)[0] for path in sounds]
        return sounds
    
    @cached_property
    def textures(self) -> list[str]:
        """
        Returns a list of all textures in the pack, relative to the pack root.

        Example: "textures/my_texture.png"
        """
        glob_pattern = os.path.join(self.input_path, "textures") + "/**/*."
        for extension in TEXTURE_EXTENSIONS:
            self._textures.extend(glob.glob(glob_pattern + extension, recursive=True)) 

        self._textures = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in self._textures]
        return self._textures
    
    def get_textures(self, search_path: str = "", trim_extension: bool = True) -> list[str]:
        """
        Returns a list of all child textures of the searchpath, relative to the pack root. 
        Search path should not include 'textures'.

        You may optionally trim the extension from the returned paths.

        Example: rp.get_textures("entities", trim_extension=True)
        """
        textures = []
        glob_pattern = os.path.join(self.input_path, "textures", search_path) + "/**/*."
        for extension in TEXTURE_EXTENSIONS:
            textures.extend(glob.glob(glob_pattern + extension, recursive=True))

        textures = [os.path.relpath(path, self.input_path).replace(os.sep, '/') for path in textures]
        if trim_extension:
            textures = [os.path.splitext(path)[0] for path in textures]
        return textures

    # === Individual Files ===

    @SingleResourceDefinition(SoundsFile)
    def sounds_file(self): pass

    @SingleResourceDefinition(SoundDefinitionsFile)
    def sound_definitions_file(self) -> SoundDefinitionsFile: pass

    @SingleResourceDefinition(FlipbookTexturesFile)
    def flipbook_textures_file(self): pass

    @SingleResourceDefinition(BlocksFile)
    def blocks_file(self): pass

    @SingleResourceDefinition(BiomesClientFile)
    def biomes_client_file(self): pass

    @SingleResourceDefinition(TerrainTextureFile)
    def terrain_texture_file(self): pass

    @SingleResourceDefinition(ItemTextureFile)
    def item_texture_file(self): pass